﻿/**
 * Valencia Renderer - realtime Mandelbulb demo
 * Copyright (C) 2013 Adam Celarek (www.celarek.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef FRACTALDISTANCEESTIMATOR_H
#define FRACTALDISTANCEESTIMATOR_H

#include "DoubleMath.h"
#include "ValenciaRenderer.cuh"

using valencia::myFloat;
using valencia::myFloat3;
using valencia::myFloat4;
using valencia::make_myFloat4;
using valencia::make_myFloat3;


struct FractalDistanceEstimator
{
  VALENCIA_INLINE __host__ __device__
  FractalDistanceEstimator(char maxIterations):m_max_iterations(maxIterations) {
  }

  __host__ __device__ __forceinline__
  myFloat abs(myFloat a) {
      if(a>0) return a;
      return -a;
  }

  // Return the approximate lower bound on the distance from x to the set.
  __host__ __device__ __forceinline__
	myFloat operator()( myFloat3 z) {
		/* * /
		// from http://blog.hvidtfeldts.net/index.php/2011/09/distance-estimated-3d-fractals-v-the-mandelbulb-different-de-approximations/
		myFloat Power = 8.0f;
		myFloat dr = 1.0f;
		myFloat r = 0.0f;
		myFloat3 startPos = z;
		if(minDist != 0)
			*minDist = startPos;

		for (unsigned int i = 0; i < m_max_iterations ; i++) {
			r = length(z);
			if (r>FRACTAL_Z_LENGTH_BAILOUT) break;
			dr =  valencia::myPow( r, Power-1.0f)*Power*dr + 1.0f;

			{
//				// convert to polar coordinates
//				myFloat theta = acos(z.z/r);
//				myFloat phi = atan2(z.y, z.x);
//
//				// scale and rotate the point
//				myFloat zr = valencia::myPow( r,Power);
//				theta = theta*Power;
//				phi = phi*Power;
//
//				// convert back to cartesian coordinates
//				z = zr * make_myFloat3(sin(theta)*cos(phi), sin(phi)*sin(theta), cos(theta));
				
				// non trig version from http://www.iquilezles.org/www/articles/mandelbulb/mandelbulb.htm
				// the website says, that it's much faster, but on the graphics card it was about 4 times slower.
				float _x = z.x; float x2 = _x*_x; float x4 = x2*x2;
				float _y = z.y; float y2 = _y*_y; float y4 = y2*y2;
				float _z = z.z; float z2 = _z*_z; float z4 = z2*z2;

				float k3 = x2 + z2;
				float k2 = rsqrtf ( k3*k3*k3*k3*k3*k3*k3 );
				float k1 = x4 + y4 + z4 - 6.0*y2*z2 - 6.0*x2*y2 + 2.0*z2*x2;
				float k4 = x2 - y2 + z2;

				z.x =  64.0*_x*_y*_z*(x2-z2)*k4*(x4-6.0*x2*z2+z4)*k1*k2;
				z.y = -16.0*y2*k3*k4*k4 + k1*k1;
				z.z = -8.0*_y*k4*(x4*x4 - 28.0*x4*x2*z2 + 70.0*x4*z4 - 28.0*x2*z2*z4 + z4*z4)*k1*k2;

			}

			z += startPos;
		}
		return 0.5f*log(r)*r/dr;
		/* */

		/* */
		// from http://blog.hvidtfeldts.net/index.php/2011/09/distance-estimated-3d-fractals-v-the-mandelbulb-different-de-approximations/
		myFloat Power = 8.0f;
		myFloat dr = 1.0f;
		myFloat r = 0.0f;
		myFloat3 startPos = z;

		for (unsigned char i = 0; i < m_max_iterations ; i++) {
			r = length(z);
			if (r>FRACTAL_Z_LENGTH_BAILOUT) break;

			// convert to polar coordinates
			myFloat theta = acos(z.z/r);
			myFloat phi = atan2(z.y, z.x);
			dr =  valencia::myPow( r, Power-1.0f)*Power*dr + 1.0f;

			// scale and rotate the point
			myFloat zr = valencia::myPow( r,Power);
			theta = theta*Power;
			phi = phi*Power;

			// convert back to cartesian coordinates
			z = zr * make_myFloat3(sin(theta)*cos(phi), sin(phi)*sin(theta), cos(theta));

			z += startPos;
		}
		return 0.5f*log(r)*r/dr;
		/* */
	  
    /* * /
    // from mandelbulb optix code
    const myFloat3 c = z;                    // constant term in iterative formula
    const myFloat p = 8;                 // power of fractal

    // divergence threshold nearly infinite for negative powers
    myFloat threshold = 1000.0;
    const myFloat d_threshold = (p < 0) ? 1E9 : myMin((myFloat) 2.0, myPow(threshold,((myFloat)1.0)/p));

    // orbit trapping continues at existing m_min_dist
    myFloat min_dst = m_min_dst;

    // point z polar coordinates
    myFloat phase1 = 0;
    myFloat phase = 0;
    myFloat R  = length(z);
    myFloat th = atan2(z.y, z.x);
    myFloat ph = asin(z.z / R) + phase1;  // initial phase offset

    // orbit trapping relative to point (0,0,0)
    if (R < min_dst) min_dst = R;

    myFloat3 dz;
    myFloat ph_dz = 0.0f;
    myFloat th_dz = 0.0f;
    myFloat R_dz  = 1.0f;

    // Iterate to compute the distance estimator.
    int i = m_max_iterations;

    while( i-- )
    {
      if (p>0)
      {
        // positive powers p

        // purely scalar dz iteration (Thanks to Enforcer for the great tip)
        myFloat zr = myPow(R, p-1);
        R_dz = zr*R_dz*p + 1;

        // z iteration
        myFloat P_ = zr * R; // equivalent to __powf(R, p);
        myFloat s1, c1; mySincos(p*ph, &s1, &c1);
        myFloat s2, c2; mySincos(p*th, &s2, &c2);
        z.x = P_ * c1*c2 + c.x;
        z.y = P_ * c1*s2 + c.y;
        z.z = P_ * s1    + c.z;
      }
      else
      {
        // negative powers p

        // derivative dz iteration
        myFloat pP_ = p*myPow(R, p-1.0f);
        myFloat c1, s1;   mySincos(ph_dz+(p-1.0f)*ph, &s1, &c1);
        myFloat s1_, c1_; mySincos(th_dz+(p-1.0f)*th, &s1_, &c1_);
        dz.x = pP_ * R_dz*c1*c1_+1.0f;
        dz.y = pP_ * R_dz*c1*s1_;
        dz.z = pP_ * R_dz*s1;

        // polar coordinates of derivative dz
        R_dz  = sqrt(dz.x*dz.x + dz.y*dz.y + dz.z*dz.z);
        th_dz = atan2(dz.y, dz.x);
        ph_dz = asin(dz.z /R_dz);

        // z iteration
        myFloat P_ = myPow(R, p);
        myFloat s2, c2; mySincos(p*ph, &s2, &c2);
        myFloat s2_, c2_; mySincos(p*th, &s2_, &c2_);
        z.x = P_ * c2*c2_ + c.x;
        z.y = P_ * c2*s2_ + c.y;
        z.z = P_ * s2    + c.z;
      }
      // compute new length of z
      R  = length(z);

      // results are not stored for the "extra" iteration at i == 0
      // orbit trapping relative to point (0,0,0)
      if (R < min_dst) min_dst = R;

      // Stop when we know the point diverges and return the result.
      if( R > d_threshold ) break;

      // compute remaining polar coordinates of z and iterate further
      th = atan2(z.y, z.x);
      ph = asin(z.z / R) + phase;  // iterative phase offset
    }

    // return the result if we reached convergence
    m_min_dst = min_dst;

    return ((myFloat)0.5) * R * log(R)/R_dz;
    /* */
  }

  unsigned char m_max_iterations;
};

#endif // FRACTALDISTANCEESTIMATOR_H
