/**
 * Valencia Renderer - realtime Mandelbulb demo
 * Copyright (C) 2013 Adam Celarek (www.celarek.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef RAY_H
#define RAY_H

#include "DoubleMath.h"
#include "Constants.h"
#include "ValenciaRenderer.cuh"

namespace valencia {

struct Ray {
	__device__ Ray(const myFloat3& origin, const myFloat3& direction) : origin(origin), direction(direction) {} 
	__device__ Ray() {}
	__device__ static Ray fromCamera(myFloat x, myFloat y, myFloat width, myFloat height);	// use given screen space coords for calculating ray
	__device__ static Ray fromCamera(unsigned char position);	// init phase: use position "offset" in addition to threadIdx and co. see bottom for more explenation
	__device__ static unsigned char positionFromCamera();		// final phase: get position within acceleration structure, that was used for the corresponding camera ray
	__device__ static Ray fromCamera();		// use threadIdx and co.
	myFloat3 origin;
	myFloat3 direction;
};

__device__ Ray Ray::fromCamera(myFloat x, myFloat y, myFloat width, myFloat height) {
	myFloat2 positionInViewSpace = make_myFloat2(x/width, y/height) * 2.0 - make_myFloat2(1.0);
	myFloat3 direction = normalize(positionInViewSpace.x * cameraData[0].U + positionInViewSpace.y * cameraData[0].V + cameraData[0].W);
	return Ray(cameraData[0].eye, direction);
}

__device__ Ray Ray::fromCamera() {
    myFloat x = blockIdx.x*blockDim.x + threadIdx.x;
    myFloat y = blockIdx.y*blockDim.y + threadIdx.y;

	myFloat width = gridDim.x * blockDim.x;
	myFloat height = gridDim.y * blockDim.y;

	return fromCamera(x, y, width, height);
}


#ifdef ACCELERATION_ACCURATE_RAY_BLOCK
/*
we are in acceleration init pass
positions start with 0 in the top left and count up as shown.
ACCELERATION_STRUCT_MIDDLE_RAY indicats the centre
 __ __ __ __ __ __ __
|_0|_1|_2|_3|_4|_5|_6|
|_7|_8|_9|10|11|12|..|
|__|__|__|__|__|__|20|
|__|__|__|24|__|__|__|
|__|__|__|__|__|__|__|
|__|__|__|__|__|__|__|
|__|__|__|__|__|__|48|

*/
__device__ Ray Ray::fromCamera(unsigned char position) {
	unsigned int deltaX = 0;
	unsigned int deltaY = 0;
	
	deltaX = position % ACCELERATION_BLOCK_SIZE;
	deltaY = position / ACCELERATION_BLOCK_SIZE;

	myFloat x = ACCELERATION_BLOCK_SIZE * (blockIdx.x*blockDim.x + threadIdx.x) + ACCELERATION_BLOCK_SIZE/2 + deltaX;
	myFloat y = ACCELERATION_BLOCK_SIZE * (blockIdx.y*blockDim.y + threadIdx.y) + ACCELERATION_BLOCK_SIZE/2 + deltaY;

	myFloat width = gridDim.x * blockDim.x * ACCELERATION_BLOCK_SIZE;
	myFloat height = gridDim.y * blockDim.y * ACCELERATION_BLOCK_SIZE;

	return fromCamera(x, y, width, height);
}

__device__ unsigned char Ray::positionFromCamera() {
	// localX/Y are 0/0 at position 0
	int localX = (blockIdx.x*blockDim.x + threadIdx.x)%ACCELERATION_BLOCK_SIZE;
	int localY = (blockIdx.y*blockDim.y + threadIdx.y)%ACCELERATION_BLOCK_SIZE;

	return ACCELERATION_BLOCK_SIZE * localY + localX;
}

#else
/*
positions start with 0 in the centre and count up as shown.
 _ _ _ _ _ _ _
|_|_|_|6|7|8|9|
|_|_|_|3|4|5|_|
|_|_|_|1|2|_|_|
|_|_|_|0|_|_|_|
|_|_|_|_|_|_|_|
|_|_|_|_|_|_|_|
|_|_|_|_|_|_|_|

we assume, that we are in acceleration init pass. the calculated rays with position > 0 are not intended for final computation as they are produced only to the top and right of the centre.
they are used only for initialising of the acceleration structure.

the highest position could be calculated easily by a arithmetical progression formula, but it's faster to use constants:
ACCELERATION_BLOCK_SIZE = 3 -> 2
ACCELERATION_BLOCK_SIZE = 5 -> 5
ACCELERATION_BLOCK_SIZE = 7 -> 9
we probably won't need more.

*/
__device__ Ray Ray::fromCamera(unsigned char position) {
	unsigned int deltaX = 0;
	unsigned int deltaY = 0;
	
	if(position == 2 || position == 4 || position == 7)
		deltaX = 1;
	if(position == 5 || position == 8)
		deltaX = 2;
	if(position == 9)
		deltaX = 3;
	
	
	if(position == 1 || position == 2)
		deltaY = 1;
	if(position > 2 && position < 6) // 3, 4, 5
		deltaY = 2;
	if(position > 5)  // 6, 7, 8, 9 -> top row
		deltaY = 3;

	myFloat x = ACCELERATION_BLOCK_SIZE * (blockIdx.x*blockDim.x + threadIdx.x) + ACCELERATION_BLOCK_SIZE/2 + deltaX;
	myFloat y = ACCELERATION_BLOCK_SIZE * (blockIdx.y*blockDim.y + threadIdx.y) + ACCELERATION_BLOCK_SIZE/2 + deltaY;

	myFloat width = gridDim.x * blockDim.x * ACCELERATION_BLOCK_SIZE;
	myFloat height = gridDim.y * blockDim.y * ACCELERATION_BLOCK_SIZE;

	return fromCamera(x, y, width, height);
}

/*
 _ _ _ _ _ _ _
|_|_|_|6|7|8|9|
|_|_|_|3|4|5|8|
|_|_|_|1|2|4|7|
|_|_|_|0|1|3|6|
|_|_|_|_|_|_|_|
|_|_|_|_|_|_|_|
|_|_|_|_|_|_|_|
*/
__device__ unsigned char Ray::positionFromCamera() {
	// localX/Y are 0/0 at position 0
	int localX = (blockIdx.x*blockDim.x + threadIdx.x)%ACCELERATION_BLOCK_SIZE - ACCELERATION_BLOCK_SIZE/2;
	int localY = (blockIdx.y*blockDim.y + threadIdx.y)%ACCELERATION_BLOCK_SIZE - ACCELERATION_BLOCK_SIZE/2;

	if(localX<0) localX = -localX;
	if(localY<0) localY = -localY;

	// ugly, but i don't have any other idea. get the position using ifs.
	if(localX <= 1 && localY <= 1)
		return localX + localY;

	if(localX == localY) {
		if(localX == 2)
			return 5;
		return 9;
	}
	if(localX == 0 || localY == 0) {
		if(localX == 2 || localY == 2)
			return 3;
		return 6;
	}

	if(localX == 1 || localY == 1) {
		if(localX == 2 || localY == 2)
			return 4;
		return 7;
	}

	return 8;
	
}

#endif


}

#endif