/**
 * Valencia Renderer - realtime Mandelbulb demo
 * Copyright (C) 2013 Adam Celarek (www.celarek.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef RAY_ACCELERATION_STRUCTURE_H
#define RAY_ACCELERATION_STRUCTURE_H

#include "Ray.h"


namespace valencia {
struct __align__(8) RayAccelerationData {
	unsigned char rayStepsForRayPosition[ACCELERATION_RAY_COUNT];
	float rayDistanceForRayPosition[ACCELERATION_RAY_COUNT];
};

__device__ RayAccelerationData rayAccelerators[RENDER_WIDTH / ACCELERATION_BLOCK_SIZE][RENDER_HEIGHT / ACCELERATION_BLOCK_SIZE];
#define RAY_ACCELERATOR_INIT_PHASE rayAccelerators[blockIdx.x*blockDim.x + threadIdx.x][blockIdx.y*blockDim.y + threadIdx.y]
#define RAY_ACCELERATOR_FINAL_PHASE rayAccelerators[(blockIdx.x*blockDim.x + threadIdx.x)/ACCELERATION_BLOCK_SIZE][(blockIdx.y*blockDim.y + threadIdx.y)/ACCELERATION_BLOCK_SIZE]


struct RayAccelerationStructure {
	__device__ RayAccelerationStructure(Ray ray) : ray(ray), availableRaySteps(0) {
		for(unsigned char i=1; i<ACCELERATION_RAY_COUNT; i++) {
			Ray acceleratedRay = Ray::fromCamera(i);

			// TODO: we have to do this computation for all ACCELERATION_BLOCK_SIZE * ACCELERATION_BLOCK_SIZE rays !!
			rayKs[i] = length(ray.direction - acceleratedRay.direction);
		}
		rayKs[0] = 0;
	}

	__device__ void addSphere(myFloat t, myFloat distance) {
		//deSpheres[availableRaySteps].x = t;
		//deSpheres[availableRaySteps].y = distance;

		if(availableRaySteps > 0) {
			//myFloat r1 = deSpheres[availableRaySteps-2].y;
			myFloat r2sq = distance * distance;
			myFloat c2 = t;
			myFloat distFromC2toIntercectionT = 0.5f * r2sq / r1;

			myFloat intersectionT = c2 - distFromC2toIntercectionT;
			myFloat intersectionRingRadius = sqrt(r2sq - distFromC2toIntercectionT * distFromC2toIntercectionT);

			for(unsigned char i=0; i<ACCELERATION_RAY_COUNT; i++) {

				// maybe it's possible to remove the if term after &&. but it would need mathematical proof.
				//     or, if doing the memory access for the last step only -> if we managed to go through both rings -> no need to check distance sphere.
				if(intersectionT * rayKs[i] < intersectionRingRadius && t * rayKs[i] < distance) {
					// the accelerated ray went through the ring and is inside the current distance sphere.
					// add the data to the acceleration structure.
					RAY_ACCELERATOR_INIT_PHASE.rayStepsForRayPosition[i] = availableRaySteps;
					RAY_ACCELERATOR_INIT_PHASE.rayDistanceForRayPosition[i] = t; // it might be faster to do the global memory access only for the last (failing) step.
				}
				else {
					// the ray is not anymore inside the ring or the distance sphere. even if it gets inside later, we shouldn't consider it.
					// a very high k will make it impossible to fall inside the ring again
					// test, if bools etc would be faster!
					rayKs[i] = 1e35f;
				}
			}

			//rings[availableRaySteps-2].x = intersectionT;
			//rings[availableRaySteps-2].y = intersectionRingRadius;
		}
		availableRaySteps++;
		r1 = distance;
	}

	__device__ static unsigned char getNeighbouringSteps(unsigned char position) {
		int localX = (blockIdx.x*blockDim.x + threadIdx.x)%ACCELERATION_BLOCK_SIZE - ACCELERATION_BLOCK_SIZE/2;
		int localY = (blockIdx.y*blockDim.y + threadIdx.y)%ACCELERATION_BLOCK_SIZE - ACCELERATION_BLOCK_SIZE/2;

		//if(localX == 0 && localY == 0)
		//	return RAY_ACCELERATOR_FINAL_PHASE.rayStepsForRayPosition[position];

		return rayAccelerators[((blockIdx.x*blockDim.x + threadIdx.x)/ACCELERATION_BLOCK_SIZE)+localX][((blockIdx.y*blockDim.y + threadIdx.y)/ACCELERATION_BLOCK_SIZE)+localY].rayStepsForRayPosition[position];
	}

	__device__ static void accelerateRay(unsigned char* steps, myFloat* distance) {
		unsigned char position = Ray::positionFromCamera();
		*steps = (RAY_ACCELERATOR_FINAL_PHASE.rayStepsForRayPosition[position]);// + getNeighbouringSteps(position))/2;
		*distance = RAY_ACCELERATOR_FINAL_PHASE.rayDistanceForRayPosition[position];
	}

	Ray ray;
	// k stands for the factor, that an acceleration ray diverges, when going along it for 1 t
	float rayKs[ACCELERATION_RAY_COUNT];
	unsigned char availableRaySteps;

	//myFloat2 deSpheres[FRACTAL_MAX_RAY_STEPS];
	myFloat r1; // previous sphere radius ( r_{n-1} )

	//myFloat2 rings[FRACTAL_MAX_RAY_STEPS];
};

}

#endif