/**
 * Valencia Renderer - realtime Mandelbulb demo
 * Copyright (C) 2013 Adam Celarek (www.celarek.at)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "ValenciaRenderer.cuh"
#include <iostream>

#include "CudaHelperFunctions.cuh"
#include "Ray.h"
#include "RayAccelerationStructure.h"
#include "DistanceEstimator.h"
#include "NormalEstimator.h"
#include "Constants.h"

using namespace valencia;

__device__ float2 defferedShadingOutData[RENDER_WIDTH*RENDER_HEIGHT];
texture<float2, cudaTextureType2D,  cudaReadModeElementType> defferedShadingTexture;
cudaArray* defferedShadingInData;


__device__ float4 doShadingMaxSteps(const myFloat3& position, myFloat distToCamera, unsigned int stepCount, const Ray& ray) {
	return make_float4(0, 0, 0, 1);
}

__device__ float calcShadow(const myFloat3& position, const myFloat3& lightDirection) {

	FractalDistanceEstimator distance(renderParams[0].fractalMaxIterations);
	myFloat3 currentPosition = position;

	myFloat distFromOrigin = distance(currentPosition);
	currentPosition += distFromOrigin * lightDirection;

	for(unsigned int i = 0; i <= renderParams[0].maxMarchingSteps; ++i )
    {
		myFloat dist = distance(currentPosition);

		// Step along the ray and accumulate the distance from the origin.
		currentPosition += dist * lightDirection;
		distFromOrigin += dist;

		// Check if we're close enough or too far.
		if(dist < fmaxf(BOARDER_OF_PRECISION, renderParams[0].shadowEpsilonAlpha * distFromOrigin)) {
			return 0.2f;
		}
		if(distFromOrigin > RAY_MAX_DISTANCE) {
			return 1.f;
		}
    }
	return 0.2f;
}

__device__ float4 doShading(const myFloat3& position, const myFloat3& normal, myFloat distToCamera, unsigned int stepCount, const Ray& ray) {
	float4 colour = make_float4(0);
	colour.w = 1.0f;
	/* */
	float fakeGI = 1 - (float) stepCount / (float) renderParams[0].maxMarchingSteps;
	/* */
	// diffuse light one
	float light1 = max(0.0, dot(normal,normalize(make_float3(1, 1, 1))));

	if(renderParams[0].shadow)
		light1 *= calcShadow(position, normalize(make_float3(1, 1, 1)) );

	// diffuse light two
	float light2 = max(0.0, dot(normal, normalize(make_float3(-1, 0, -1))))* 0.3;

	// diffuse light three
	float light3 = max(0.0, dot(normal, normalize(make_float3(0, -1, 1))))* 0.3;
	

	/*
	FractalDistanceEstimator distance(FRACTAL_MAX_ITERATIONS-2);
	myFloat3 minDistColour;
	distance(position, &minDistColour);

	colour.x = light * minDistColour.x;
	colour.y = light * minDistColour.y;
	colour.z = light * minDistColour.z;
	*/
	/* */
	colour.x = (light1*1.0f + light2*0.8 + light3*0.6) * fakeGI;
	colour.y = (light1*0.9f + light2*0.6 + light3*0.6) * fakeGI;
	colour.z = (light1*0.8f + light2*0.6 + light3*0.9) * fakeGI;
	/* * /
	
	colour.x = fakeGI;
	colour.y = fakeGI;
	colour.z = fakeGI;

	/* * /
	colour.x *= normal.x;
	colour.y *= normal.y;
	colour.z *= normal.z;

	if(colour.x < 0)
		colour.x *= -1.f;
	if(colour.y < 0)
		colour.y *= -1.f;
	if(colour.z < 0)
		colour.z *= -1.f;

	/* */
	return colour;
}

__device__ float4 doBackgroundShading(const Ray& ray) {
	float4 colour = make_float4(0);
	colour.w = 1.0f;
	colour.x = 0.f;
	colour.y = 0.f;
	colour.z = 0.f;
	
	return colour;
}


__global__  void
fractalComputationTraditionalDE(unsigned int *od)
{
	Ray ray = Ray::fromCamera();
	FractalDistanceEstimator distance(renderParams[0].fractalMaxIterations);
	myFloat distFromOrigin = 0;

	float2 defferedData;

	myFloat3 currentPosition = ray.origin;
    for(unsigned int i = 0; i <= renderParams[0].maxMarchingSteps; ++i )
    {
		myFloat dist = distance(currentPosition);

		// Step along the ray and accumulate the distance from the origin.
		currentPosition += dist * ray.direction;
		distFromOrigin += dist;

		// Check if we're close enough or too far.
		if(dist < fmaxf(BOARDER_OF_PRECISION, renderParams[0].fractalEpsilonAlpha * distFromOrigin) || i == renderParams[0].maxMarchingSteps) {
			defferedData.x = distFromOrigin;
			defferedData.y = i;
			break;
		}
		if(distFromOrigin > RAY_MAX_DISTANCE) {
			defferedData.x = distFromOrigin;
			defferedData.y = i;
			break;
		}
    }

    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;
	unsigned int width = gridDim.x * blockDim.x;
	defferedShadingOutData[x+y*width] = defferedData;
}

__global__  void
fractalComputationAccelerationStructureInit(unsigned int *od)
{
	Ray ray = Ray::fromCamera(ACCELERATION_STRUCT_MIDDLE_RAY);
	RayAccelerationStructure accelerator(ray);

	FractalDistanceEstimator distance(renderParams[0].fractalMaxIterations);
	myFloat distFromOrigin = 0;

	//float4 colour = make_float4(0);

	myFloat3 currentPosition = ray.origin;
	for(unsigned int i = 0; i <= renderParams[0].maxMarchingSteps; ++i )
    {
		myFloat dist = distance(currentPosition);

		// Step along the ray and accumulate the distance from the origin.
		currentPosition += dist * ray.direction;
		distFromOrigin += dist;

		accelerator.addSphere(distFromOrigin, dist);
		//accelerator.availableRaySteps++;
		//rayAccelerators[blockIdx.x*blockDim.x + threadIdx.x][blockIdx.y*blockDim.y + threadIdx.y].availableRaySteps++;

		// Check if we're close enough or too far.
		if(dist < fmaxf(BOARDER_OF_PRECISION, renderParams[0].fractalEpsilonAlpha * distFromOrigin)) {
			break;
		}
		if(distFromOrigin > RAY_MAX_DISTANCE) {
			break;
		}
    }

	//rayAccelerators[blockIdx.x*blockDim.x + threadIdx.x][blockIdx.y*blockDim.y + threadIdx.y] = accelerator;

}

__global__  void
fractalComputationCompute(unsigned int *od)
{
	Ray ray = Ray::fromCamera();

	FractalDistanceEstimator distance(renderParams[0].fractalMaxIterations);
	myFloat distFromOrigin = 0;

	
	unsigned char steps;
	RayAccelerationStructure::accelerateRay(&steps, &distFromOrigin);

	//float4 colour = make_float4(1 - distFromOrigin * 0.5f);
	//float4 colour = make_float4((float)steps/100.f);
	//float4 colour;
	float2 defferedData;

	myFloat3 currentPosition = ray.origin + distFromOrigin * ray.direction;
	for(unsigned int i = steps; i <= renderParams[0].maxMarchingSteps; ++i )
	{
		myFloat dist = distance(currentPosition);

		// Step along the ray and accumulate the distance from the origin.
		currentPosition += dist * ray.direction;
		distFromOrigin += dist;

		// Check if we're close enough or too far.
		if(dist < fmaxf(BOARDER_OF_PRECISION, renderParams[0].fractalEpsilonAlpha * distFromOrigin) || i == renderParams[0].maxMarchingSteps) {
			//myFloat3 normal = make_myFloat3(estimate_normal(distance, currentPosition));
			//colour = doShading(currentPosition, normal, distFromOrigin, i, ray);
			defferedData.x = distFromOrigin;
			defferedData.y = i;
			break;
		}
		if(distFromOrigin > RAY_MAX_DISTANCE) {
			//colour = doBackgroundShading(ray);
			defferedData.x = distFromOrigin;
			defferedData.y = i;
			break;
		}
	}

    unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;
	unsigned int width = gridDim.x * blockDim.x;
	//od[x+y*width] = rgbaFloatToInt(colour);
	defferedShadingOutData[x+y*width] = defferedData;

}

__global__ void doDefferedRender(unsigned int *out) {
	unsigned int x = blockIdx.x*blockDim.x + threadIdx.x;
    unsigned int y = blockIdx.y*blockDim.y + threadIdx.y;
	unsigned int width = gridDim.x * blockDim.x;

	Ray ray = Ray::fromCamera();
	float4 colour;
	float distFromOrigin = defferedShadingOutData[x+y*width].x;

	if(distFromOrigin < RAY_MAX_DISTANCE) {
		unsigned int steps = defferedShadingOutData[x+y*width].y;
		myFloat3 currentPosition = ray.origin + distFromOrigin * ray.direction;
		if(steps < renderParams[0].maxMarchingSteps) {
			FractalDistanceEstimator distance(renderParams[0].fractalMaxIterations);
			myFloat3 normal = make_myFloat3(estimate_normal(distance, currentPosition));
			colour = doShading(currentPosition, normal, distFromOrigin, steps, ray);
		}
		else {
			colour = doShadingMaxSteps(currentPosition, distFromOrigin, steps, ray);
		}
	}
	else {
		colour = doBackgroundShading(ray);
	}

	out[x+y*width] = rgbaFloatToInt(colour);
}

//ValenciaRenderer::ValenciaRenderer(int width, int height) : m_camera(make_myFloat3(1.057, 0, 0.14), make_myFloat3(0, 0, 1), make_myFloat3(0.5, 2, 0)), m_width(width), m_height(height) {
ValenciaRenderer::ValenciaRenderer(int width, int height) : m_camera(make_myFloat3(2, 0, 0), make_myFloat3(0, 0, 0), make_myFloat3(0, 1, 0)), m_width(width), m_height(height) {
	updateCamera();
	m_timer = 0;
	sdkCreateTimer(&m_timer);
	sdkStartTimer(&m_timer);
	m_acceleratedRenderMode = true;

	m_renderParams.shadow = true;
	m_renderParams.fractalEpsilonAlpha = FRACTAL_EPSILON_ALPHA;
	m_renderParams.fractalMaxIterations = FRACTAL_MAX_ITERATIONS;
	m_renderParams.maxMarchingSteps = FRACTAL_MAX_RAY_STEPS;
	m_renderParams.shadowEpsilonAlpha = SHADOW_EPSILON_ALPHA;
	updateRenderParam();

	//cudaMallocPitch()
	
    //checkCudaErrors(cudaMalloc((void **) &d_temp, (RENDER_WIDTH * RENDER_HEIGHT * sizeof(unsigned int))));

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float2>();
	checkCudaErrors(cudaMallocArray(&defferedShadingInData, &channelDesc, RENDER_WIDTH, RENDER_HEIGHT));
}

ValenciaRenderer::~ValenciaRenderer() {
	sdkDeleteTimer(&m_timer);
}

double ValenciaRenderer::render(unsigned int *d_dest, StopWatchInterface *timer)
{
    // var for kernel computation timing
    double initKernelTime;
	double computeKernelTime;

    // sync host and start kernel computation timer_kernel
	initKernelTime = computeKernelTime = 0.0;
    sdkResetTimer(&timer);
	sdkStartTimer(&timer);

	if(m_acceleratedRenderMode) {
		dim3 threadsPerBlockInit(THREADS_PER_BLOCK_WIDTH, THREADS_PER_BLOCK_HEIGHT);
		dim3 numBlocksInit((m_width/ACCELERATION_BLOCK_SIZE) / threadsPerBlockInit.x, (m_height/ACCELERATION_BLOCK_SIZE) / threadsPerBlockInit.y);
		fractalComputationAccelerationStructureInit<<< numBlocksInit, threadsPerBlockInit, 0 >>>(d_dest);
		
		checkCudaErrors(cudaDeviceSynchronize());
		//sdkResetTimer(&timer);
		initKernelTime = sdkGetTimerValue(&timer);
	
		dim3 threadsPerBlock(THREADS_PER_BLOCK_WIDTH, THREADS_PER_BLOCK_HEIGHT);
		dim3 numBlocks(m_width / threadsPerBlock.x, m_height / threadsPerBlock.y);
		fractalComputationCompute<<< numBlocks, threadsPerBlock, 0 >>>(d_dest);

		
	}
	else {
		dim3 threadsPerBlock(THREADS_PER_BLOCK_WIDTH, THREADS_PER_BLOCK_HEIGHT);
		dim3 numBlocks(m_width / threadsPerBlock.x, m_height / threadsPerBlock.y);
		fractalComputationTraditionalDE<<< numBlocks, threadsPerBlock, 0 >>>(d_dest);

		//doDefferedRender<<< numBlocks, threadsPerBlock, 0 >>>(d_dest);
	}
	
	dim3 threadsPerBlock(THREADS_PER_BLOCK_WIDTH, THREADS_PER_BLOCK_HEIGHT);
	dim3 numBlocks(m_width / threadsPerBlock.x, m_height / threadsPerBlock.y);
	checkCudaErrors(cudaDeviceSynchronize());
	//sdkResetTimer(&timer);
	//initKernelTime = sdkGetTimerValue(&timer);

	doDefferedRender<<< numBlocks, threadsPerBlock, 0 >>>(d_dest);

    // sync host and stop computation timer_kernel
    checkCudaErrors(cudaDeviceSynchronize());
    computeKernelTime = sdkGetTimerValue(&timer);
	sdkStopTimer(&timer);

	m_deltaTime = sdkGetTimerValue(&m_timer);
	sdkResetTimer(&m_timer);

    //return computeKernelTime/1000.f;
	return m_deltaTime;//computeKernelTime/(computeKernelTime);
}


void ValenciaRenderer::moveForward() {
	FractalDistanceEstimator distance(m_renderParams.fractalMaxIterations);
	myFloat distanceToFractal = distance(m_camera.getCameraData()->eye);
	m_camera.translate(distanceToFractal * 0.001 * m_deltaTime);

	updateCamera();
}

void ValenciaRenderer::moveBackward() {
	FractalDistanceEstimator distance(m_renderParams.fractalMaxIterations);
	myFloat distanceToFractal = distance(m_camera.getCameraData()->eye);
	m_camera.translate(distanceToFractal * -0.001 * m_deltaTime);

	updateCamera();
}

void ValenciaRenderer::rotateCamera(float dx, float dy) {
	m_camera.rotate(make_myFloat2(dx, dy));
	updateCamera();
}

void ValenciaRenderer::updateCamera() {
	cudaMemcpyToSymbol(cameraData, m_camera.getCameraData(), sizeof(PinholeCamera::CameraData));
}

void ValenciaRenderer::updateRenderParam() {
	cudaMemcpyToSymbol(renderParams, &m_renderParams, sizeof(RenderParameters));
}

void ValenciaRenderer::printInfos() {
	FractalDistanceEstimator distance(m_renderParams.fractalMaxIterations);
	myFloat distanceToFractal = distance(m_camera.getCameraData()->eye);

	std::cout << "distance to fractal: " << distanceToFractal << std::endl;
}

bool ValenciaRenderer::toggleShadows() {
	m_renderParams.shadow = !m_renderParams.shadow;
	updateRenderParam();
	return m_renderParams.shadow;
}

char ValenciaRenderer::changFractalMaxIterations(Direction dir) {
	m_renderParams.fractalMaxIterations += (dir==Increase)?1:(-1);
	updateRenderParam();
	return m_renderParams.fractalMaxIterations;
}

short ValenciaRenderer::changMaxMarchingSteps(Direction dir) {
	m_renderParams.maxMarchingSteps += (dir==Increase)?5:(-5);
	updateRenderParam();
	return m_renderParams.maxMarchingSteps;
}

float ValenciaRenderer::changFractalEpsilonAlpha(Direction dir) {
	m_renderParams.fractalEpsilonAlpha *= (dir==Increase)?1.1f:(1.f/1.1f);
	updateRenderParam();
	return m_renderParams.fractalEpsilonAlpha;
}

float ValenciaRenderer::changShadowEpsilonAlpha(Direction dir) {
	m_renderParams.shadowEpsilonAlpha *= (dir==Increase)?1.1f:(1.f/1.1f);
	updateRenderParam();
	return m_renderParams.shadowEpsilonAlpha;
}

